/*
 *  Login.app
 * 
 *  Copyright (c) 1997-2000 Per Liden
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include <stdio.h>
#include <string.h>
#include "util.h"
#include "preferences.h"
#include "config.h"

static char* illegalValueMsg(char* entryName) {
   char* msg = "illegal value assigned to \"%s\"";
   char*       buffer;

   buffer = alloc(strlen(msg) + strlen(entryName));
   sprintf(buffer, msg, entryName);
   return buffer;
}

static Bool readString(WMPropList *pl, char* entryName, char** buffer, char* defaultValue) {
   WMPropList *entry;
   WMPropList *key;

   key = WMCreatePLString(entryName);
   entry = WMGetFromPLDictionary(pl, key);
   WMReleasePropList(key);
   if (!entry) {
      /* Use default if not found */
      *buffer = strdup(defaultValue);
   } else if (!WMIsPLString(entry)) {
      /* Not a string */
      return False;
   } else {	
      /* Use proplist value */
      *buffer = strdup(WMGetFromPLString(entry));
      WMReleasePropList(entry);
   }

   return True;
}

static Bool
readStringArray(WMPropList *pl, char* entryName, char*** buffer, char* defaultValue)
{
   char* space = " ";
   WMPropList * entry;
   WMPropList * key;
   char*       value;
   char*       tmp1;
   char*       tmp2;
   int         i;
   int         numOfStrings;
    
   key = WMCreatePLString((char*)entryName);
   entry = WMGetFromPLDictionary(pl, key);
   WMReleasePropList(key);
   if (!entry) {
      /* Use default if not found */
      value = strdup(defaultValue);
   } else if (!WMIsPLString(entry)) {
      /* Not a string */
      return False;
   } else {	
      /* Use proplist value */
      value = strdup(WMGetFromPLString(entry));
      WMReleasePropList(entry);
   }

   /* Count number of tokens */
   tmp1 = strdup(value);
   tmp2 = strtok(tmp1, space);    
   for (numOfStrings = 0; tmp2 != NULL; numOfStrings++) {
      tmp2 = strtok(NULL, space);
   }
   numOfStrings++;
   free(tmp2);

   /* Insert tokens into array */
   *buffer = alloc(sizeof(char*) * numOfStrings);
   (*buffer)[0] = strtok(value, space);    
   for (i = 1; i < numOfStrings; i++) {
      (*buffer)[i] = strtok(NULL, space);
   }

   return True;
}

static Bool
readBoolean(WMPropList *pl, char* entryName, Bool* buffer, Bool defaultValue)
{
   WMPropList *entry;
   WMPropList *key;

   key = WMCreatePLString((char*)entryName);
   entry = WMGetFromPLDictionary(pl, key);
   WMReleasePropList(key);
   if (!entry) {
      /* Use default if not found */
      *buffer = defaultValue;
   } else if (!WMIsPLString(entry)) {
      /* Not a string */
      return False;
   } else {	
      /* Use proplist value */
      *buffer = !strcasecmp(WMGetFromPLString(entry), "YES")? True : False;
      WMReleasePropList(entry);
   }

   return True;
}

static Bool
readInteger(WMPropList *pl, char* entryName, int* buffer, int defaultValue)
{
   WMPropList *entry;
   WMPropList *key;

   key = WMCreatePLString((char*)entryName);
   entry = WMGetFromPLDictionary(pl, key);
   WMReleasePropList(key);
   if (!entry) {
      /* Use default if not found */
      *buffer = defaultValue;
   } else if (!WMIsPLString(entry)) {
      /* Not a string */
      return False;
   } else {	
      /* Use proplist value */
      sscanf(WMGetFromPLString(entry), "%d", buffer);
      WMReleasePropList(entry);
   }

   return True;
}

static Bool
readPosition(WMPropList *pl, char* entryName, int** buffer, int defaultValueX, int defaultValueY)
{
   WMPropList *entry;
   WMPropList *key;

   key = WMCreatePLString((char*)entryName);
   entry = WMGetFromPLDictionary(pl, key);
   WMReleasePropList(key);
   if (!entry) {
      /* Use default if not found */
      (*buffer)[0] = defaultValueX;
      (*buffer)[1] = defaultValueY;
   } else if (!WMIsPLString(entry)) {
      /* Not a string */
      return False;
   } else {	
      /* Use proplist value */
      sscanf(WMGetFromPLString(entry), "%d, %d", (int*)&buffer[0], (int*)&(buffer[1]));
      WMReleasePropList(entry);
   }

   return True;
}

static Bool
readJustification(WMPropList *pl, char* entryName, int* buffer, int defaultValue)
{
   WMPropList *entry;
   WMPropList *key;

   key = WMCreatePLString((char*)entryName);
   entry = WMGetFromPLDictionary(pl, key);
   WMReleasePropList(key);
   if (!entry) {
      /* Use default if not found */
      *buffer = defaultValue;
   } else if (!WMIsPLString(entry)) {
      /* Not a string */
      return False;
   } else {	
      /* Use proplist value */
      char* value = WMGetFromPLString(entry);
      if (!strcasecmp(value, "RIGHT")) {
         *buffer = JUSTIFICATION_RIGHT;
      } else if (!strcasecmp(value, "LEFT")) {
         *buffer = JUSTIFICATION_LEFT;
      } else {
         *buffer = JUSTIFICATION_CENTER;
      }
      WMReleasePropList(entry);
   }

   return True;
}

static char*
readPreferences(char* file, Preferences* pref) {
   char* entryWelcomeMessage = "WelcomeMessage";
   char* entryRestartMessage = "RestartMessage";
   char* entryPowerOffMessage = "PowerOffMessage";
   char* entryXServer = "XServer";
   char* entryTheme = "Theme";
   char* entryDisableWelcomeMessage = "DisableWelcomeMessage";
   char* entryDisableUserShutdown = "DisableUserShutdown";
   char* entryDisableHostname = "DisableHostname";
   char* entryDisableAnimations = "DisableAnimations";
   WMPropList * pl;

   /* Read file */
   pl = WMReadPropListFromFile(file);
   if (pl == NULL) {
      return "could not read preferences";
   }

   /* WelcomeMessage */
   if (readString(pl, entryWelcomeMessage, &pref->welcomeMessage, DEFAULT_WELCOMEMESSAGE) == False) {
      return illegalValueMsg(entryWelcomeMessage);
   }

   /* RestartMessage */
   if (readString(pl, entryRestartMessage, &pref->restartMessage, DEFAULT_RESTARTMESSAGE) == False) {
      return illegalValueMsg(entryRestartMessage);
   }

   /* PowerOffMessage */
   if (readString(pl, entryPowerOffMessage, &pref->powerOffMessage, DEFAULT_POWEROFFMESSAGE) == False) {
      return illegalValueMsg(entryPowerOffMessage);
   }

   /* XServer */
   if (readStringArray(pl, entryXServer, &pref->xserver, DEFAULT_XSERVER) == False) {
      return illegalValueMsg(entryXServer);
   }

   /* Theme */
   if (readString(pl, entryTheme, &pref->theme, DEFAULT_THEME) == False) {
      return illegalValueMsg(entryTheme);
   }

   /* DisableWelcomeMessage */
   if (readBoolean(pl, entryDisableWelcomeMessage, &pref->disableWelcomeMessage, DEFAULT_DISABLEWELCOMEMESSAGE) == False) {
      return illegalValueMsg(entryDisableWelcomeMessage);
   }

   /* DisableUserShutdown */
   if (readBoolean(pl, entryDisableUserShutdown, &pref->disableUserShutdown, DEFAULT_DISABLEUSERSHUTDOWN) == False) {
      return illegalValueMsg(entryDisableUserShutdown);
   }

   /* DisableHostname */
   if (readBoolean(pl, entryDisableHostname, &pref->disableHostname, DEFAULT_DISABLEHOSTNAME) == False) {
      return illegalValueMsg(entryDisableHostname);
   }

   /* DisableAnimations */
   if (readBoolean(pl, entryDisableAnimations, &pref->disableAnimations, DEFAULT_DISABLEANIMATIONS) == False) {
      return illegalValueMsg(entryDisableAnimations);
   }

   /* WMReleasePropList(pl); */
   return NULL;
}

static char*
readThemeSettings(char* file, Preferences* pref) {
   char* entryPanelImage = "PanelImage";
   char* entryFieldLabelsImage = "FieldLabelsImage";
   char* entryPowerButtonUpImage = "PowerButtonUpImage";
   char* entryPowerButtonDownImage = "PowerButtonDownImage";
   char* entryRestartButtonUpImage = "RestartButtonUpImage";
   char* entryRestartButtonDownImage = "RestartButtonDownImage";
   char* entryAlertIcon = "AlertIcon";
   char* entryBackgroundColor = "BackgroundColor";
   char* entryPanelPosition = "PanelPosition";
   char* entryPowerButtonPosition = "PowerButtonPosition";
   char* entryRestartButtonPosition = "RestartButtonPosition";
   char* entryMessagePosition = "MessagePosition";
   char* entryMessageJustification = "MessageJustification";
   char* entryMessageColor = "MessageColor";
   char* entryMessageFont = "MessageFont";
   char* entryHostnamePosition = "HostnamePosition";
   char* entryHostnameJustification = "HostnameJustification";
   char* entryHostnameColor = "HostnameColor";
   char* entryHostnameFont = "HostnameFont";
   char* entryInputFont = "InputFont";
   char* entryInputFontColor = "InputFontColor";
   char* entryInputCursorHeight = "InputCursorHeight";
   char* entryInputUserImage = "InputUserImage";
   char* entryInputUserPosition = "InputUserPosition";
   char* entryInputPasswordImage = "InputPasswordImage";
   char* entryInputPasswordPosition = "InputPasswordPosition";
   WMPropList * pl;

   /* Read file */
   pl = WMReadPropListFromFile(file);
   if (pl == NULL) {
      return "could not read theme settings";
   }

   /* PanelImage */
   if (readString(pl, entryPanelImage, &pref->panelImage, DEFAULT_PANELIMAGE) == False) {
      return illegalValueMsg(entryPanelImage);
   }

   /* FieldLabelsImage */
   if (readString(pl, entryFieldLabelsImage, &pref->fieldLabelsImage, DEFAULT_FIELDLABELSIMAGE) == False) {
      return illegalValueMsg(entryFieldLabelsImage);
   }

   /* PowerButtonUpImage */
   if (readString(pl, entryPowerButtonUpImage, &pref->powerButtonUpImage, DEFAULT_POWERBUTTONUPIMAGE) == False) {
      return illegalValueMsg(entryPowerButtonUpImage);
   }

   /* PowerButtonDownImage */
   if (readString(pl, entryPowerButtonDownImage, &pref->powerButtonDownImage, DEFAULT_POWERBUTTONDOWNIMAGE) == False) {
      return illegalValueMsg(entryPowerButtonDownImage);
   }

   /* RestartButtonUpImage */
   if (readString(pl, entryRestartButtonUpImage, &pref->restartButtonUpImage, DEFAULT_RESTARTBUTTONUPIMAGE) == False) {
      return illegalValueMsg(entryRestartButtonUpImage);
   }

   /* RestartButtonDownImage */
   if (readString(pl, entryRestartButtonDownImage, &pref->restartButtonDownImage, DEFAULT_RESTARTBUTTONDOWNIMAGE) == False) {
      return illegalValueMsg(entryRestartButtonDownImage);
   }

   /* AlertIcon */
   if (readString(pl, entryAlertIcon, &pref->alertIcon, DEFAULT_ALERTICON) == False) {
      return illegalValueMsg(entryAlertIcon);
   }

   /* BackgroundColor */
   if (readString(pl, entryBackgroundColor, &pref->backgroundColor, DEFAULT_BACKGROUNDCOLOR) == False) {
      return illegalValueMsg(entryBackgroundColor);
   }

   /* PanelPosition */
   if (readPosition(pl, entryPanelPosition, (int**)&pref->panelPosition, DEFAULT_PANELPOSITION_X, DEFAULT_PANELPOSITION_Y) == False) {
      return illegalValueMsg(entryPanelPosition);
   }

   /* PowerButtonPosition */
   if (readPosition(pl, entryPowerButtonPosition, (int**)&pref->powerButtonPosition, DEFAULT_POWERBUTTONPOSITION_X, DEFAULT_POWERBUTTONPOSITION_Y) == False) {
      return illegalValueMsg(entryPowerButtonPosition);
   }

   /* RestartButtonPosition */
   if (readPosition(pl, entryRestartButtonPosition, (int**)&pref->restartButtonPosition, DEFAULT_RESTARTBUTTONPOSITION_X, DEFAULT_RESTARTBUTTONPOSITION_Y) == False) {
      return illegalValueMsg(entryRestartButtonPosition);
   }

   /* MessagePosition */
   if (readPosition(pl, entryMessagePosition, (int**)&pref->messagePosition, DEFAULT_MESSAGEPOSITION_X, DEFAULT_MESSAGEPOSITION_Y) == False) {
      return illegalValueMsg(entryMessagePosition);
   }

   /* MessageJustification */
   if (readJustification(pl, entryMessageJustification, &pref->messageJustification, DEFAULT_MESSAGEJUSTIFICATION) == False) {
      return illegalValueMsg(entryMessageJustification);
   }

   /* MessageFont */
   if (readString(pl, entryMessageFont, &pref->messageFont, DEFAULT_MESSAGEFONT) == False) {
      return illegalValueMsg(entryMessageFont);
   }

   /* MessageColor */
   if (readString(pl, entryMessageColor, &pref->messageColor, DEFAULT_MESSAGECOLOR) == False) {
      return illegalValueMsg(entryMessageColor);
   }

   /* HostnamePosition */
   if (readPosition(pl, entryHostnamePosition, (int**)&pref->hostnamePosition, DEFAULT_HOSTNAMEPOSITION_X, DEFAULT_HOSTNAMEPOSITION_Y) == False) {
      return illegalValueMsg(entryHostnamePosition);
   }

   /* HostnameJustification */
   if (readJustification(pl, entryHostnameJustification, &pref->hostnameJustification, DEFAULT_HOSTNAMEJUSTIFICATION) == False) {
      return illegalValueMsg(entryHostnameJustification);
   }

   /* HostnameFont */
   if (readString(pl, entryHostnameFont, &pref->hostnameFont, DEFAULT_HOSTNAMEFONT) == False) {
      return illegalValueMsg(entryHostnameFont);
   }

   /* HostnameColor */
   if (readString(pl, entryHostnameColor, &pref->hostnameColor, DEFAULT_HOSTNAMECOLOR) == False) {
      return illegalValueMsg(entryHostnameColor);
   }

   /* InputFont */
   if (readString(pl, entryInputFont, &pref->inputFont, DEFAULT_INPUTFONT) == False) {
      return illegalValueMsg(entryInputFont);
   }

   /* InputFontColor */
   if (readString(pl, entryInputFontColor, &pref->inputFontColor, DEFAULT_INPUTFONTCOLOR) == False) {
      return illegalValueMsg(entryInputFontColor);
   }

   /* InputCursorHeight */
   if (readInteger(pl, entryInputCursorHeight, &pref->inputCursorHeight, DEFAULT_INPUTCURSORHEIGHT) == False) {
      return illegalValueMsg(entryInputCursorHeight);
   }

   /* InputUserImage */
   if (readString(pl, entryInputUserImage, &pref->inputUserImage, DEFAULT_INPUTUSERIMAGE) == False) {
      return illegalValueMsg(entryInputUserImage);
   }

   /* InputUserPosition */
   if (readPosition(pl, entryInputUserPosition, (int**)&pref->inputUserPosition, DEFAULT_INPUTUSERPOSITION_X, DEFAULT_INPUTUSERPOSITION_Y) == False) {
      return illegalValueMsg(entryInputUserPosition);
   }

   /* InputPasswordImage */
   if (readString(pl, entryInputPasswordImage, &pref->inputPasswordImage, DEFAULT_INPUTPASSWORDIMAGE) == False) {
      return illegalValueMsg(entryInputPasswordImage);
   }

   /* InputPasswordPosition */
   if (readPosition(pl, entryInputPasswordPosition, (int**)&pref->inputPasswordPosition, DEFAULT_INPUTPASSWORDPOSITION_X, DEFAULT_INPUTPASSWORDPOSITION_Y) == False) {
      return illegalValueMsg(entryInputPasswordPosition);
   }

   /* WMReleasePropList(pl); */
   return NULL;
}

Preferences* readConfiguration(char* programPath) {
   Preferences* pref;
   char*        path;
   char*        file;
   char*  errorMsg;
   char*  errorText = "%s\n\nYou can check your configuration with the \"--checkconfig\" option\nPress ENTER to terminate Login.app\n";

   /* Initialize preferences structure */
   pref = alloc(sizeof(Preferences));

   /* Get program path */
   path = getApplicationPath(programPath);

   /* Read preferences */
   file = stringCombine(path, FILE_PREFERENCES);
   errorMsg = readPreferences(file, pref);
   free(file);

   if (errorMsg != NULL) {    
      printError(errorText, errorMsg);
      getchar();
      exit(1);
   }

   /* Read theme settings */
   file = stringCombine(stringCombine(stringCombine(path, THEME_DIR"/"), pref->theme), "/"FILE_THEME_SETTINGS);
   errorMsg = readThemeSettings(file, pref);
   free(file);
    
   if (errorMsg != NULL) {
      printError(errorText, errorMsg);
      getchar();
      exit(1);
   }

   free(path);
   return pref;
}

void checkConfiguration(char* programPath) {
   Preferences* pref;
   char*  yes = "Yes";
   char*  no = "No";
   char*  errorMsg;
   char*        path;
   char*        file;

   /* Initialize preferences structure */
   pref = alloc(sizeof(Preferences));

   /* Get program path */
   path = getApplicationPath(programPath);

   /* Read preferences */
   file = stringCombine(path, FILE_PREFERENCES);
   printf("Preferences: (file \"%s\")\n", file);
   errorMsg = readPreferences(file, pref);
   free(file);

   if (errorMsg != NULL) {
      free(path);
      printError("%s\n", errorMsg);
      return;
   } else {
      int i;
      printf("  WelcomeMessage:\t\t\"%s\"\n", pref->welcomeMessage);
      printf("  RestartMessage:\t\t\"%s\"\n", pref->restartMessage);
      printf("  PowerOffMessage:\t\t\"%s\"\n", pref->powerOffMessage);
      printf("  XServer:\t\t\t\"");
      for(i = 0; pref->xserver[i] != NULL; i++) {
         printf("%s", pref->xserver[i]);
         if (pref->xserver[i+1] != NULL) {
            printf(" ");
         }
      }	
      printf("\"\n");
      printf("  Theme:\t\t\t\"%s\"\n", pref->theme);
      printf("  DisableWelcomeMessage:\t%s\n", pref->disableWelcomeMessage == True ? yes : no);
      printf("  DisableUserShutdown:\t\t%s\n", pref->disableUserShutdown == True ? yes : no);
      printf("  DisableHostname:\t\t%s\n", pref->disableHostname == True ? yes : no);
      printf("  DisableAnimations:\t\t%s\n", pref->disableAnimations == True ? yes : no);
   }

   /* Read theme settings */
   file = stringCombine(stringCombine(stringCombine(path, THEME_DIR"/"), pref->theme), "/"FILE_THEME_SETTINGS);
   printf("\nTheme settings: (file \"%s\")\n", file);
   errorMsg = readThemeSettings(file, pref);
   free(file);
    
   if (errorMsg != NULL) {
      free(path);
      printError("%s\n", errorMsg);
      return;
   } else {
      printf("  PanelImage:\t\t\t\"%s\"\n", pref->panelImage);
      printf("  FieldLabelsImage:\t\t\"%s\"\n", pref->fieldLabelsImage);
      printf("  PowerButtonUpImage:\t\t\"%s\"\n", pref->powerButtonUpImage);
      printf("  PowerButtonDownImage:\t\t\"%s\"\n", pref->powerButtonDownImage);
      printf("  RestartButtonUpImage:\t\t\"%s\"\n", pref->restartButtonUpImage);
      printf("  RestartButtonDownImage:\t\"%s\"\n", pref->restartButtonDownImage);
      printf("  AlertIcon:\t\t\t\"%s\"\n", pref->alertIcon);
      printf("  BackgroundColor:\t\t\"%s\"\n", pref->backgroundColor);
      printf("  PanelPosition:\t\t%d, %d\n", pref->panelPosition[0], pref->panelPosition[1]);
      printf("  PowerButtonPosition:\t\t%d, %d\n", pref->powerButtonPosition[0], pref->powerButtonPosition[1]);
      printf("  RestartButtonPosition:\t%d, %d\n", pref->restartButtonPosition[0], pref->restartButtonPosition[1]);
      printf("  MessagePosition:\t\t%d, %d\n", pref->messagePosition[0], pref->messagePosition[1]);
      printf("  MessageJustification:\t\t");
      if (pref->messageJustification == JUSTIFICATION_LEFT) {
         printf("Left\n");
      } else if (pref->messageJustification == JUSTIFICATION_RIGHT) {
         printf("Right\n");
      } else {
         printf("Center\n");
      }
      printf("  MessageColor:\t\t\t\"%s\"\n", pref->messageColor);
      printf("  MessageFont:\t\t\t\"%s\"\n", pref->messageFont);
      printf("  HostnamePosition:\t\t%d, %d\n", pref->hostnamePosition[0], pref->hostnamePosition[1]);
      printf("  HostnameJustification:\t");
      if (pref->hostnameJustification == JUSTIFICATION_LEFT) {
         printf("Left\n");
      } else if (pref->hostnameJustification == JUSTIFICATION_RIGHT) {
         printf("Right\n");
      } else {
         printf("Center\n");
      }
      printf("  HostnameColor:\t\t\"%s\"\n", pref->hostnameColor);
      printf("  HostnameFont:\t\t\t\"%s\"\n", pref->hostnameFont);
      printf("  InputFont:\t\t\t\"%s\"\n", pref->inputFont);
      printf("  InputFontColor:\t\t\"%s\"\n", pref->inputFontColor);
      printf("  InputCursorHeight:\t\t%d\n", pref->inputCursorHeight);
      printf("  InputUserImage:\t\t%s\n", pref->inputUserImage);
      printf("  InputUserPosition:\t\t%d, %d\n", pref->inputUserPosition[0], pref->inputUserPosition[1]);
      printf("  InputPasswordImage:\t\t%s\n", pref->inputPasswordImage);
      printf("  InputPasswordPosition:\t%d, %d\n", pref->inputPasswordPosition[0], pref->inputPasswordPosition[1]);
   }

   free(path);
   free(pref);
}

